'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.classWithPrivateMethods = classWithPrivateMethods;
exports.extendClassWithPrivateMethods = extendClassWithPrivateMethods;
exports.privateMethod = privateMethod;

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var _privates = [];

var BASE_OBJECT_PROTOTYPE = Object.prototype;

function classWithPrivateMethods(target) {
  var ClassWrapper = function ClassWrapper() {
    _classCallCheck(this, ClassWrapper);

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var instance = new (Function.prototype.bind.apply(target, [null].concat(args)))();

    Object.defineProperties(this, {
      __origInstance: { value: instance },
      __classWithPrivateMethodsMethodMap: { value: {} }
    });
  };

  hoistPublicMethods(ClassWrapper, target);
  hoistStaticMethods(ClassWrapper, target);

  Object.defineProperties(ClassWrapper, {
    name: { value: 'ClassWithPrivateMethodsWrapper(' + target.name + ')' },
    __origClass: { value: target, enumerable: false, configurable: true },
    __origInstance: { value: target, enumerable: false, configurable: true },
    __classWithPrivateMethodsMethodMap: { value: {} }
  });

  return ClassWrapper;
}

function hoistPublicMethods(ClassWrapper, target) {
  var proto = target.prototype;
  var methodNames = Object.getOwnPropertyNames(proto);

  // add all methods from base class, if any
  var baseClassPrototype = Object.getPrototypeOf(proto);
  if (baseClassPrototype !== BASE_OBJECT_PROTOTYPE) {
    methodNames = methodNames.concat(Object.getOwnPropertyNames(Object.getPrototypeOf(proto)));
  }

  methodNames = methodNames.filter(function (m) {
    return m !== 'constructor';
  });

  var privatesForTarget = new Set(_privates.filter(function (e) {
    return e.target === target;
  }).map(function (e) {
    return e.name;
  }));

  methodNames.filter(function (methodName) {
    return !privatesForTarget.has(methodName);
  }).forEach(function (methodName) {
    return addBoundMethod(ClassWrapper.prototype, methodName, proto);
  });
}

function hoistStaticMethods(ClassWrapper, target) {
  var statics = Object.getOwnPropertyNames(target).filter(function (k) {
    return k !== 'length' && k !== 'name' && k !== 'prototype';
  });

  // static methods - bind methods to original class
  statics.filter(function (k) {
    return typeof target[k] === 'function';
  }).forEach(function (methodName) {
    return addBoundMethod(ClassWrapper, methodName, target);
  });

  // static properties - create accessors
  statics.filter(function (k) {
    return typeof target[k] !== 'function';
  }).forEach(function (k) {
    return Object.defineProperty(ClassWrapper, k, {
      get: function get() {
        return target[k];
      },
      set: function set(v) {
        return target[k] = v;
      }
    });
  });
}

// adds a property to *target* with:
// - a memoized getter that returns a bound method to the current instance
// - a setter that sets both wrapper and instance's method
function addBoundMethod(target, methodName, methodsContainer) {
  Object.defineProperty(target, methodName, {
    configurable: true,
    get: function get() {
      // direct read from Class.prototype.method needs to return the method itself
      if (!this.__classWithPrivateMethodsMethodMap) return methodsContainer[methodName];
      // reads from an instance need to return a bound method to the instance
      return this.__classWithPrivateMethodsMethodMap[methodName] || (this.__classWithPrivateMethodsMethodMap[methodName] = methodsContainer[methodName].bind(this.__origInstance));
    },
    set: function set(newFn) {
      if (!this.__classWithPrivateMethodsMethodMap) return methodsContainer[methodName] = newFn;
      this.__classWithPrivateMethodsMethodMap[methodName] = newFn.bind(this.__origInstance);
      this.__origInstance[methodName] = newFn;
    }
  });
}

function extendClassWithPrivateMethods(target, obj) {
  obj = Object.assign({}, obj);
  Object.assign(target.__origClass.prototype, obj);

  var objProperties = Object.getOwnPropertyNames(obj);

  objProperties.filter(function (k) {
    return typeof obj[k] === 'function';
  }).forEach(function (k) {
    return addBoundMethod(target.prototype, k, obj);
  });

  objProperties.filter(function (k) {
    return typeof obj[k] !== 'function';
  }).forEach(function (k) {
    return target.prototype[k] = obj[k];
  });
}

function privateMethod(target, name) {
  _privates.push({ target: target.constructor, name: name });
}